import pytest

from authlib.common.encoding import base64_to_int
from authlib.common.encoding import json_dumps
from authlib.jose import ECKey
from authlib.jose import JsonWebKey
from authlib.jose import KeySet
from authlib.jose import OctKey
from authlib.jose import OKPKey
from authlib.jose import RSAKey
from tests.util import read_file_path


def test_oct_import_oct_key():
    # https://tools.ietf.org/html/rfc7520#section-3.5
    obj = {
        "kty": "oct",
        "kid": "018c0ae5-4d9b-471b-bfd6-eef314bc7037",
        "use": "sig",
        "alg": "HS256",
        "k": "hJtXIZ2uSN5kbQfbtTNWbpdmhkV8FJG-Onbc6mxCcYg",
    }
    key = OctKey.import_key(obj)
    new_obj = key.as_dict()
    assert obj["k"] == new_obj["k"]
    assert "use" in new_obj


def test_oct_invalid_oct_key():
    with pytest.raises(ValueError):
        OctKey.import_key({})


def test_oct_generate_oct_key():
    with pytest.raises(ValueError):
        OctKey.generate_key(251)

    with pytest.raises(ValueError, match="oct key can not be generated as public"):
        OctKey.generate_key(is_private=False)

    key = OctKey.generate_key()
    assert "kid" in key.as_dict()
    assert "use" not in key.as_dict()

    key2 = OctKey.import_key(key, {"use": "sig"})
    assert "use" in key2.as_dict()


def test_rsa_import_ssh_pem():
    raw = read_file_path("ssh_public.pem")
    key = RSAKey.import_key(raw)
    obj = key.as_dict()
    assert obj["kty"] == "RSA"


def test_rsa_public_key():
    # https://tools.ietf.org/html/rfc7520#section-3.3
    obj = read_file_path("jwk_public.json")
    key = RSAKey.import_key(obj)
    new_obj = key.as_dict()
    assert base64_to_int(new_obj["n"]) == base64_to_int(obj["n"])
    assert base64_to_int(new_obj["e"]) == base64_to_int(obj["e"])


def test_rsa_private_key():
    # https://tools.ietf.org/html/rfc7520#section-3.4
    obj = read_file_path("jwk_private.json")
    key = RSAKey.import_key(obj)
    new_obj = key.as_dict(is_private=True)
    assert base64_to_int(new_obj["n"]) == base64_to_int(obj["n"])
    assert base64_to_int(new_obj["e"]) == base64_to_int(obj["e"])
    assert base64_to_int(new_obj["d"]) == base64_to_int(obj["d"])
    assert base64_to_int(new_obj["p"]) == base64_to_int(obj["p"])
    assert base64_to_int(new_obj["q"]) == base64_to_int(obj["q"])
    assert base64_to_int(new_obj["dp"]) == base64_to_int(obj["dp"])
    assert base64_to_int(new_obj["dq"]) == base64_to_int(obj["dq"])
    assert base64_to_int(new_obj["qi"]) == base64_to_int(obj["qi"])


def test_rsa_private_key2():
    rsa_obj = read_file_path("jwk_private.json")
    obj = {
        "kty": "RSA",
        "kid": "bilbo.baggins@hobbiton.example",
        "use": "sig",
        "n": rsa_obj["n"],
        "d": rsa_obj["d"],
        "e": "AQAB",
    }
    key = RSAKey.import_key(obj)
    new_obj = key.as_dict(is_private=True)
    assert base64_to_int(new_obj["n"]) == base64_to_int(obj["n"])
    assert base64_to_int(new_obj["e"]) == base64_to_int(obj["e"])
    assert base64_to_int(new_obj["d"]) == base64_to_int(obj["d"])
    assert base64_to_int(new_obj["p"]) == base64_to_int(rsa_obj["p"])
    assert base64_to_int(new_obj["q"]) == base64_to_int(rsa_obj["q"])
    assert base64_to_int(new_obj["dp"]) == base64_to_int(rsa_obj["dp"])
    assert base64_to_int(new_obj["dq"]) == base64_to_int(rsa_obj["dq"])
    assert base64_to_int(new_obj["qi"]) == base64_to_int(rsa_obj["qi"])


def test_invalid_rsa():
    with pytest.raises(ValueError):
        RSAKey.import_key({"kty": "RSA"})
    rsa_obj = read_file_path("jwk_private.json")
    obj = {
        "kty": "RSA",
        "kid": "bilbo.baggins@hobbiton.example",
        "use": "sig",
        "n": rsa_obj["n"],
        "d": rsa_obj["d"],
        "p": rsa_obj["p"],
        "e": "AQAB",
    }
    with pytest.raises(ValueError):
        RSAKey.import_key(obj)


def test_rsa_key_generate():
    with pytest.raises(ValueError):
        RSAKey.generate_key(256)
    with pytest.raises(ValueError):
        RSAKey.generate_key(2001)

    key1 = RSAKey.generate_key(is_private=True)
    assert b"PRIVATE" in key1.as_pem(is_private=True)
    assert b"PUBLIC" in key1.as_pem(is_private=False)

    key2 = RSAKey.generate_key(is_private=False)
    with pytest.raises(ValueError):
        key2.as_pem(True)
    assert b"PUBLIC" in key2.as_pem(is_private=False)


def test_ec_public_key():
    # https://tools.ietf.org/html/rfc7520#section-3.1
    obj = read_file_path("secp521r1-public.json")
    key = ECKey.import_key(obj)
    new_obj = key.as_dict()
    assert new_obj["crv"] == obj["crv"]
    assert base64_to_int(new_obj["x"]) == base64_to_int(obj["x"])
    assert base64_to_int(new_obj["y"]) == base64_to_int(obj["y"])
    assert key.as_json()[0] == "{"


def test_ec_private_key():
    # https://tools.ietf.org/html/rfc7520#section-3.2
    obj = read_file_path("secp521r1-private.json")
    key = ECKey.import_key(obj)
    new_obj = key.as_dict(is_private=True)
    assert new_obj["crv"] == obj["crv"]
    assert base64_to_int(new_obj["x"]) == base64_to_int(obj["x"])
    assert base64_to_int(new_obj["y"]) == base64_to_int(obj["y"])
    assert base64_to_int(new_obj["d"]) == base64_to_int(obj["d"])


def test_invalid_ec():
    with pytest.raises(ValueError):
        ECKey.import_key({"kty": "EC"})


def test_ec_key_generate():
    with pytest.raises(ValueError):
        ECKey.generate_key("Invalid")

    key1 = ECKey.generate_key("P-384", is_private=True)
    assert b"PRIVATE" in key1.as_pem(is_private=True)
    assert b"PUBLIC" in key1.as_pem(is_private=False)

    key2 = ECKey.generate_key("P-256", is_private=False)
    with pytest.raises(ValueError):
        key2.as_pem(True)
    assert b"PUBLIC" in key2.as_pem(is_private=False)


def test_import_okp_ssh_key():
    raw = read_file_path("ed25519-ssh.pub")
    key = OKPKey.import_key(raw)
    obj = key.as_dict()
    assert obj["kty"] == "OKP"
    assert obj["crv"] == "Ed25519"


def test_import_okp_public_key():
    obj = {
        "x": "AD9E0JYnpV-OxZbd8aN1t4z71Vtf6JcJC7TYHT0HDbg",
        "crv": "Ed25519",
        "kty": "OKP",
    }
    key = OKPKey.import_key(obj)
    new_obj = key.as_dict()
    assert obj["x"] == new_obj["x"]


def test_import_okp_private_pem():
    raw = read_file_path("ed25519-pkcs8.pem")
    key = OKPKey.import_key(raw)
    obj = key.as_dict(is_private=True)
    assert obj["kty"] == "OKP"
    assert obj["crv"] == "Ed25519"
    assert "d" in obj


def test_import_okp_private_dict():
    obj = {
        "x": "11qYAYKxCrfVS_7TyWQHOg7hcvPapiMlrwIaaPcHURo",
        "d": "nWGxne_9WmC6hEr0kuwsxERJxWl7MmkZcDusAxyuf2A",
        "crv": "Ed25519",
        "kty": "OKP",
    }
    key = OKPKey.import_key(obj)
    new_obj = key.as_dict(is_private=True)
    assert obj["d"] == new_obj["d"]


def test_okp_key_generate_pem():
    with pytest.raises(ValueError):
        OKPKey.generate_key("invalid")

    key1 = OKPKey.generate_key("Ed25519", is_private=True)
    assert b"PRIVATE" in key1.as_pem(is_private=True)
    assert b"PUBLIC" in key1.as_pem(is_private=False)

    key2 = OKPKey.generate_key("X25519", is_private=False)
    with pytest.raises(ValueError):
        key2.as_pem(True)
    assert b"PUBLIC" in key2.as_pem(is_private=False)


def test_jwk_generate_keys():
    key = JsonWebKey.generate_key(kty="oct", crv_or_size=256, is_private=True)
    assert key["kty"] == "oct"

    key = JsonWebKey.generate_key(kty="EC", crv_or_size="P-256")
    assert key["kty"] == "EC"

    key = JsonWebKey.generate_key(kty="RSA", crv_or_size=2048)
    assert key["kty"] == "RSA"

    key = JsonWebKey.generate_key(kty="OKP", crv_or_size="Ed25519")
    assert key["kty"] == "OKP"


def test_jwk_import_keys():
    rsa_pub_pem = read_file_path("rsa_public.pem")
    with pytest.raises(ValueError):
        JsonWebKey.import_key(rsa_pub_pem, {"kty": "EC"})

    key = JsonWebKey.import_key(raw=rsa_pub_pem, options={"kty": "RSA"})
    assert "e" in dict(key)
    assert "n" in dict(key)

    key = JsonWebKey.import_key(raw=rsa_pub_pem)
    assert "e" in dict(key)
    assert "n" in dict(key)


def test_jwk_import_key_set():
    jwks_public = read_file_path("jwks_public.json")
    key_set1 = JsonWebKey.import_key_set(jwks_public)
    key1 = key_set1.find_by_kid("abc")
    assert key1["e"] == "AQAB"

    key_set2 = JsonWebKey.import_key_set(jwks_public["keys"])
    key2 = key_set2.find_by_kid("abc")
    assert key2["e"] == "AQAB"

    key_set3 = JsonWebKey.import_key_set(json_dumps(jwks_public))
    key3 = key_set3.find_by_kid("abc")
    assert key3["e"] == "AQAB"

    with pytest.raises(ValueError):
        JsonWebKey.import_key_set("invalid")


def test_jwk_find_by_kid_with_use():
    key1 = OctKey.import_key("secret", {"kid": "abc", "use": "sig"})
    key2 = OctKey.import_key("secret", {"kid": "abc", "use": "enc"})

    key_set = KeySet([key1, key2])
    key = key_set.find_by_kid("abc", use="sig")
    assert key == key1

    key = key_set.find_by_kid("abc", use="enc")
    assert key == key2


def test_jwk_find_by_kid_with_alg():
    key1 = OctKey.import_key("secret", {"kid": "abc", "alg": "HS256"})
    key2 = OctKey.import_key("secret", {"kid": "abc", "alg": "dir"})

    key_set = KeySet([key1, key2])
    key = key_set.find_by_kid("abc", alg="HS256")
    assert key == key1

    key = key_set.find_by_kid("abc", alg="dir")
    assert key == key2


def test_jwk_thumbprint():
    # https://tools.ietf.org/html/rfc7638#section-3.1
    data = read_file_path("thumbprint_example.json")
    key = JsonWebKey.import_key(data)
    expected = "NzbLsXh8uDCcd-6MNwXF4W_7noWXFZAfHkxZsRGC9Xs"
    assert key.thumbprint() == expected


def test_jwk_key_set():
    key = RSAKey.generate_key(is_private=True)
    key_set = KeySet([key])
    obj = key_set.as_dict()["keys"][0]
    assert "kid" in obj
    assert key_set.as_json()[0] == "{"



"""authlib.oauth2.rfc7662.
~~~~~~~~~~~~~~~~~~~~~~

This module represents a direct implementation of
OAuth 2.0 Token Introspection.

https://tools.ietf.org/html/rfc7662
"""

from .introspection import IntrospectionEndpoint
from .models import IntrospectionToken
from .token_validator import IntrospectTokenValidator

__all__ = ["IntrospectionEndpoint", "IntrospectionToken", "IntrospectTokenValidator"]



from authlib.jose import JsonWebKey
from authlib.jose import JsonWebToken
from authlib.jose import jwt
from authlib.oidc.core import CodeIDToken
from authlib.oidc.core import ImplicitIDToken
from authlib.oidc.core import UserInfo


class OpenIDMixin:
    def fetch_jwk_set(self, force=False):
        metadata = self.load_server_metadata()
        jwk_set = metadata.get("jwks")
        if jwk_set and not force:
            return jwk_set

        uri = metadata.get("jwks_uri")
        if not uri:
            raise RuntimeError('Missing "jwks_uri" in metadata')

        with self.client_cls(**self.client_kwargs) as session:
            resp = session.request("GET", uri, withhold_token=True)
            resp.raise_for_status()
            jwk_set = resp.json()

        self.server_metadata["jwks"] = jwk_set
        return jwk_set

    def userinfo(self, **kwargs):
        """Fetch user info from ``userinfo_endpoint``."""
        metadata = self.load_server_metadata()
        resp = self.get(metadata["userinfo_endpoint"], **kwargs)
        resp.raise_for_status()
        data = resp.json()
        return UserInfo(data)

    def parse_id_token(
        self, token, nonce, claims_options=None, claims_cls=None, leeway=120
    ):
        """Return an instance of UserInfo from token's ``id_token``."""
        if "id_token" not in token:
            return None

        load_key = self.create_load_key()

        claims_params = dict(
            nonce=nonce,
            client_id=self.client_id,
        )

        if claims_cls is None:
            if "access_token" in token:
                claims_params["access_token"] = token["access_token"]
                claims_cls = CodeIDToken
            else:
                claims_cls = ImplicitIDToken

        metadata = self.load_server_metadata()
        if claims_options is None and "issuer" in metadata:
            claims_options = {"iss": {"values": [metadata["issuer"]]}}

        alg_values = metadata.get("id_token_signing_alg_values_supported")
        if alg_values:
            _jwt = JsonWebToken(alg_values)
        else:
            _jwt = jwt

        claims = _jwt.decode(
            token["id_token"],
            key=load_key,
            claims_cls=claims_cls,
            claims_options=claims_options,
            claims_params=claims_params,
        )
        # https://github.com/authlib/authlib/issues/259
        if claims.get("nonce_supported") is False:
            claims.params["nonce"] = None

        claims.validate(leeway=leeway)
        return UserInfo(claims)

    def create_load_key(self):
        def load_key(header, _):
            jwk_set = JsonWebKey.import_key_set(self.fetch_jwk_set())
            try:
                return jwk_set.find_by_kid(
                    header.get("kid"), use="sig", alg=header.get("alg")
                )
            except ValueError:
                # re-try with new jwk set
                jwk_set = JsonWebKey.import_key_set(self.fetch_jwk_set(force=True))
                return jwk_set.find_by_kid(
                    header.get("kid"), use="sig", alg=header.get("alg")
                )

        return load_key



from flask import jsonify
from flask import request
from flask_sqlalchemy import SQLAlchemy

from authlib.common.urls import url_encode
from authlib.integrations.flask_oauth1 import AuthorizationServer
from authlib.integrations.flask_oauth1 import ResourceProtector
from authlib.integrations.flask_oauth1 import (
    create_exists_nonce_func as create_cache_exists_nonce_func,
)
from authlib.integrations.flask_oauth1 import current_credential
from authlib.integrations.flask_oauth1 import register_nonce_hooks
from authlib.integrations.flask_oauth1 import register_temporary_credential_hooks
from authlib.oauth1 import ClientMixin
from authlib.oauth1 import TemporaryCredentialMixin
from authlib.oauth1 import TokenCredentialMixin
from authlib.oauth1.errors import OAuth1Error
from tests.util import read_file_path

from ..cache import SimpleCache

db = SQLAlchemy()


class User(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    username = db.Column(db.String(40), unique=True, nullable=False)

    def get_user_id(self):
        return self.id


class Client(ClientMixin, db.Model):
    id = db.Column(db.Integer, primary_key=True)
    client_id = db.Column(db.String(48), index=True)
    client_secret = db.Column(db.String(120), nullable=False)
    default_redirect_uri = db.Column(db.Text, nullable=False, default="")
    user_id = db.Column(db.Integer, db.ForeignKey("user.id", ondelete="CASCADE"))
    user = db.relationship("User")

    def get_default_redirect_uri(self):
        return self.default_redirect_uri

    def get_client_secret(self):
        return self.client_secret

    def get_rsa_public_key(self):
        return read_file_path("rsa_public.pem")


class TokenCredential(TokenCredentialMixin, db.Model):
    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(db.Integer, db.ForeignKey("user.id", ondelete="CASCADE"))
    user = db.relationship("User")
    client_id = db.Column(db.String(48), index=True)
    oauth_token = db.Column(db.String(84), unique=True, index=True)
    oauth_token_secret = db.Column(db.String(84))

    def get_oauth_token(self):
        return self.oauth_token

    def get_oauth_token_secret(self):
        return self.oauth_token_secret


class TemporaryCredential(TemporaryCredentialMixin, db.Model):
    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(db.Integer, db.ForeignKey("user.id", ondelete="CASCADE"))
    user = db.relationship("User")
    client_id = db.Column(db.String(48), index=True)
    oauth_token = db.Column(db.String(84), unique=True, index=True)
    oauth_token_secret = db.Column(db.String(84))
    oauth_verifier = db.Column(db.String(84))
    oauth_callback = db.Column(db.Text, default="")

    def get_user_id(self):
        return self.user_id

    def get_client_id(self):
        return self.client_id

    def get_redirect_uri(self):
        return self.oauth_callback

    def check_verifier(self, verifier):
        return self.oauth_verifier == verifier

    def get_oauth_token(self):
        return self.oauth_token

    def get_oauth_token_secret(self):
        return self.oauth_token_secret


class TimestampNonce(db.Model):
    __table_args__ = (
        db.UniqueConstraint(
            "client_id", "timestamp", "nonce", "oauth_token", name="unique_nonce"
        ),
    )
    id = db.Column(db.Integer, primary_key=True)
    client_id = db.Column(db.String(48), nullable=False)
    timestamp = db.Column(db.Integer, nullable=False)
    nonce = db.Column(db.String(48), nullable=False)
    oauth_token = db.Column(db.String(84))


def exists_nonce(nonce, timestamp, client_id, oauth_token):
    q = TimestampNonce.query.filter_by(
        nonce=nonce,
        timestamp=timestamp,
        client_id=client_id,
    )
    if oauth_token:
        q = q.filter_by(oauth_token=oauth_token)
    rv = q.first()
    if rv:
        return True

    item = TimestampNonce(
        nonce=nonce,
        timestamp=timestamp,
        client_id=client_id,
        oauth_token=oauth_token,
    )
    db.session.add(item)
    db.session.commit()
    return False


def create_temporary_credential(token, client_id, redirect_uri):
    item = TemporaryCredential(
        client_id=client_id,
        oauth_token=token["oauth_token"],
        oauth_token_secret=token["oauth_token_secret"],
        oauth_callback=redirect_uri,
    )
    db.session.add(item)
    db.session.commit()
    return item


def get_temporary_credential(oauth_token):
    return TemporaryCredential.query.filter_by(oauth_token=oauth_token).first()


def delete_temporary_credential(oauth_token):
    q = TemporaryCredential.query.filter_by(oauth_token=oauth_token)
    q.delete(synchronize_session=False)
    db.session.commit()


def create_authorization_verifier(credential, grant_user, verifier):
    credential.user_id = grant_user.id  # assuming your end user model has `.id`
    credential.oauth_verifier = verifier
    db.session.add(credential)
    db.session.commit()
    return credential


def create_token_credential(token, temporary_credential):
    credential = TokenCredential(
        oauth_token=token["oauth_token"],
        oauth_token_secret=token["oauth_token_secret"],
        client_id=temporary_credential.get_client_id(),
    )
    credential.user_id = temporary_credential.get_user_id()
    db.session.add(credential)
    db.session.commit()
    return credential


def create_authorization_server(app, use_cache=False, lazy=False):
    def query_client(client_id):
        return Client.query.filter_by(client_id=client_id).first()

    if lazy:
        server = AuthorizationServer()
        server.init_app(app, query_client)
    else:
        server = AuthorizationServer(app, query_client=query_client)
    if use_cache:
        cache = SimpleCache()
        register_nonce_hooks(server, cache)
        register_temporary_credential_hooks(server, cache)
        server.register_hook("create_token_credential", create_token_credential)
    else:
        server.register_hook("exists_nonce", exists_nonce)
        server.register_hook("create_temporary_credential", create_temporary_credential)
        server.register_hook("get_temporary_credential", get_temporary_credential)
        server.register_hook("delete_temporary_credential", delete_temporary_credential)
        server.register_hook(
            "create_authorization_verifier", create_authorization_verifier
        )
        server.register_hook("create_token_credential", create_token_credential)

    @app.route("/oauth/initiate", methods=["GET", "POST"])
    def initiate():
        return server.create_temporary_credentials_response()

    @app.route("/oauth/authorize", methods=["GET", "POST"])
    def authorize():
        if request.method == "GET":
            try:
                server.check_authorization_request()
                return "ok"
            except OAuth1Error:
                return "error"
        user_id = request.form.get("user_id")
        if user_id:
            grant_user = db.session.get(User, int(user_id))
        else:
            grant_user = None
        try:
            return server.create_authorization_response(grant_user=grant_user)
        except OAuth1Error as error:
            return url_encode(error.get_body())

    @app.route("/oauth/token", methods=["POST"])
    def issue_token():
        return server.create_token_response()

    return server


def create_resource_server(app, use_cache=False, lazy=False):
    if use_cache:
        cache = SimpleCache()
        exists_nonce = create_cache_exists_nonce_func(cache)
    else:

        def exists_nonce(nonce, timestamp, client_id, oauth_token):
            q = db.session.query(TimestampNonce.nonce).filter_by(
                nonce=nonce,
                timestamp=timestamp,
                client_id=client_id,
            )
            if oauth_token:
                q = q.filter_by(oauth_token=oauth_token)
            rv = q.first()
            if rv:
                return True

            tn = TimestampNonce(
                nonce=nonce,
                timestamp=timestamp,
                client_id=client_id,
                oauth_token=oauth_token,
            )
            db.session.add(tn)
            db.session.commit()
            return False

    def query_client(client_id):
        return Client.query.filter_by(client_id=client_id).first()

    def query_token(client_id, oauth_token):
        return TokenCredential.query.filter_by(
            client_id=client_id, oauth_token=oauth_token
        ).first()

    if lazy:
        require_oauth = ResourceProtector()
        require_oauth.init_app(app, query_client, query_token, exists_nonce)
    else:
        require_oauth = ResourceProtector(app, query_client, query_token, exists_nonce)

    @app.route("/user")
    @require_oauth()
    def user_profile():
        user = current_credential.user
        return jsonify(id=user.id, username=user.username)



import pytest

from authlib.common.urls import url_decode
from authlib.common.urls import urlparse
from authlib.jose import jwt
from authlib.oauth2.rfc6749.grants import (
    AuthorizationCodeGrant as _AuthorizationCodeGrant,
)
from authlib.oauth2.rfc9068 import JWTBearerTokenGenerator
from tests.util import read_file_path

from ..models import CodeGrantMixin
from ..models import User
from ..models import save_authorization_code

issuer = "https://authlib.test/"


@pytest.fixture
def user(db):
    user = User(username="foo")
    db.session.add(user)
    db.session.commit()
    yield user
    db.session.delete(user)


@pytest.fixture
def client(client, db):
    client.set_client_metadata(
        {
            "scope": "profile",
            "redirect_uris": ["https://client.test/authorized"],
            "response_types": ["code"],
            "token_endpoint_auth_method": "client_secret_post",
            "grant_types": ["authorization_code"],
        }
    )
    db.session.add(client)
    db.session.commit()
    return client


@pytest.fixture
def jwks():
    return read_file_path("jwks_private.json")


@pytest.fixture(autouse=True)
def server(server):
    server.register_grant(AuthorizationCodeGrant)
    return server


@pytest.fixture(autouse=True)
def token_generator(server, jwks):
    class MyJWTBearerTokenGenerator(JWTBearerTokenGenerator):
        def get_jwks(self):
            return jwks

    token_generator = MyJWTBearerTokenGenerator(issuer=issuer)
    server.register_token_generator("default", token_generator)
    return token_generator


class AuthorizationCodeGrant(CodeGrantMixin, _AuthorizationCodeGrant):
    TOKEN_ENDPOINT_AUTH_METHODS = ["client_secret_basic", "client_secret_post", "none"]

    def save_authorization_code(self, code, request):
        return save_authorization_code(code, request)


def test_generate_jwt_access_token(test_client, client, user, jwks):
    res = test_client.post(
        "/oauth/authorize",
        data={
            "response_type": client.response_types[0],
            "client_id": client.client_id,
            "redirect_uri": client.redirect_uris[0],
            "scope": client.scope,
            "user_id": user.id,
        },
    )

    params = dict(url_decode(urlparse.urlparse(res.location).query))
    code = params["code"]
    res = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "authorization_code",
            "code": code,
            "client_id": client.client_id,
            "client_secret": client.client_secret,
            "scope": " ".join(client.scope),
            "redirect_uri": client.redirect_uris[0],
        },
    )

    access_token = res.json["access_token"]
    claims = jwt.decode(access_token, jwks)

    assert claims["iss"] == issuer
    assert claims["sub"] == user.id
    assert claims["scope"] == client.scope
    assert claims["client_id"] == client.client_id

    # This specification registers the 'application/at+jwt' media type, which can
    # be used to indicate that the content is a JWT access token. JWT access tokens
    # MUST include this media type in the 'typ' header parameter to explicitly
    # declare that the JWT represents an access token complying with this profile.
    # Per the definition of 'typ' in Section 4.1.9 of [RFC7515], it is RECOMMENDED
    # that the 'application/' prefix be omitted. Therefore, the 'typ' value used
    # SHOULD be 'at+jwt'.

    assert claims.header["typ"] == "at+jwt"


def test_generate_jwt_access_token_extra_claims(
    test_client, token_generator, user, client, jwks
):
    """Authorization servers MAY return arbitrary attributes not defined in any
    existing specification, as long as the corresponding claim names are collision
    resistant or the access tokens are meant to be used only within a private
    subsystem. Please refer to Sections 4.2 and 4.3 of [RFC7519] for details.
    """

    def get_extra_claims(client, grant_type, user, scope):
        return {"username": user.username}

    token_generator.get_extra_claims = get_extra_claims

    res = test_client.post(
        "/oauth/authorize",
        data={
            "response_type": client.response_types[0],
            "client_id": client.client_id,
            "redirect_uri": client.redirect_uris[0],
            "scope": client.scope,
            "user_id": user.id,
        },
    )

    params = dict(url_decode(urlparse.urlparse(res.location).query))
    code = params["code"]
    res = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "authorization_code",
            "code": code,
            "client_id": client.client_id,
            "client_secret": client.client_secret,
            "scope": " ".join(client.scope),
            "redirect_uri": client.redirect_uris[0],
        },
    )

    access_token = res.json["access_token"]
    claims = jwt.decode(access_token, jwks)
    assert claims["username"] == user.username


@pytest.mark.skip
def test_generate_jwt_access_token_no_user(test_client, client, user, jwks):
    res = test_client.post(
        "/oauth/authorize",
        data={
            "response_type": client.response_types[0],
            "client_id": client.client_id,
            "redirect_uri": client.redirect_uris[0],
            "scope": client.scope,
            #'user_id': user.id,
        },
    )

    params = dict(url_decode(urlparse.urlparse(res.location).query))
    code = params["code"]
    res = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "authorization_code",
            "code": code,
            "client_id": client.client_id,
            "client_secret": client.client_secret,
            "scope": " ".join(client.scope),
            "redirect_uri": client.redirect_uris[0],
        },
    )

    access_token = res.json["access_token"]
    claims = jwt.decode(access_token, jwks)

    assert claims["sub"] == client.client_id


def test_optional_fields(test_client, token_generator, user, client, jwks):
    token_generator.get_auth_time = lambda *args: 1234
    token_generator.get_amr = lambda *args: "amr"
    token_generator.get_acr = lambda *args: "acr"

    res = test_client.post(
        "/oauth/authorize",
        data={
            "response_type": client.response_types[0],
            "client_id": client.client_id,
            "redirect_uri": client.redirect_uris[0],
            "scope": client.scope,
            "user_id": user.id,
        },
    )

    params = dict(url_decode(urlparse.urlparse(res.location).query))
    code = params["code"]
    res = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "authorization_code",
            "code": code,
            "client_id": client.client_id,
            "client_secret": client.client_secret,
            "scope": " ".join(client.scope),
            "redirect_uri": client.redirect_uris[0],
        },
    )

    access_token = res.json["access_token"]
    claims = jwt.decode(access_token, jwks)

    assert claims["auth_time"] == 1234
    assert claims["amr"] == "amr"
    assert claims["acr"] == "acr"



import base64
import json
import struct


def to_bytes(x, charset="utf-8", errors="strict"):
    if x is None:
        return None
    if isinstance(x, bytes):
        return x
    if isinstance(x, str):
        return x.encode(charset, errors)
    if isinstance(x, (int, float)):
        return str(x).encode(charset, errors)
    return bytes(x)


def to_unicode(x, charset="utf-8", errors="strict"):
    if x is None or isinstance(x, str):
        return x
    if isinstance(x, bytes):
        return x.decode(charset, errors)
    return str(x)


def to_native(x, encoding="ascii"):
    if isinstance(x, str):
        return x
    return x.decode(encoding)


def json_loads(s):
    return json.loads(s)


def json_dumps(data, ensure_ascii=False):
    return json.dumps(data, ensure_ascii=ensure_ascii, separators=(",", ":"))


def urlsafe_b64decode(s):
    s += b"=" * (-len(s) % 4)
    return base64.urlsafe_b64decode(s)


def urlsafe_b64encode(s):
    return base64.urlsafe_b64encode(s).rstrip(b"=")


def base64_to_int(s):
    data = urlsafe_b64decode(to_bytes(s, charset="ascii"))
    buf = struct.unpack(f"{len(data)}B", data)
    return int("".join([f"{byte:02x}" for byte in buf]), 16)


def int_to_base64(num):
    if num < 0:
        raise ValueError("Must be a positive integer")

    s = num.to_bytes((num.bit_length() + 7) // 8, "big", signed=False)
    return to_unicode(urlsafe_b64encode(s))


def json_b64encode(text):
    if isinstance(text, dict):
        text = json_dumps(text)
    return urlsafe_b64encode(to_bytes(text))



import json
import time

import pytest

from authlib.oauth2.rfc6749.grants import RefreshTokenGrant as _RefreshTokenGrant

from .models import Client
from .models import OAuth2Token
from .oauth2_server import create_basic_auth


@pytest.fixture(autouse=True)
def server(server):
    class RefreshTokenGrant(_RefreshTokenGrant):
        def authenticate_refresh_token(self, refresh_token):
            try:
                item = OAuth2Token.objects.get(refresh_token=refresh_token)
                if item.is_refresh_token_active():
                    return item
            except OAuth2Token.DoesNotExist:
                return None

        def authenticate_user(self, credential):
            return credential.user

        def revoke_old_credential(self, credential):
            now = int(time.time())
            credential.access_token_revoked_at = now
            credential.refresh_token_revoked_at = now
            credential.save()
            return credential

    server.register_grant(RefreshTokenGrant)
    return server


@pytest.fixture(autouse=True)
def client(user):
    client = Client(
        user_id=user.pk,
        client_id="client-id",
        client_secret="client-secret",
        scope="",
        grant_type="refresh_token",
        token_endpoint_auth_method="client_secret_basic",
        default_redirect_uri="https://client.test",
    )
    client.save()
    yield client
    client.delete()


def test_invalid_client(factory, server):
    request = factory.post(
        "/oauth/token",
        data={"grant_type": "refresh_token", "refresh_token": "foo"},
    )
    resp = server.create_token_response(request)
    assert resp.status_code == 401
    data = json.loads(resp.content)
    assert data["error"] == "invalid_client"

    request = factory.post(
        "/oauth/token",
        data={"grant_type": "refresh_token", "refresh_token": "foo"},
        HTTP_AUTHORIZATION=create_basic_auth("invalid", "client-secret"),
    )
    resp = server.create_token_response(request)
    assert resp.status_code == 401
    data = json.loads(resp.content)
    assert data["error"] == "invalid_client"


def test_invalid_refresh_token(factory, server):
    auth_header = create_basic_auth("client-id", "client-secret")
    request = factory.post(
        "/oauth/token",
        data={"grant_type": "refresh_token"},
        HTTP_AUTHORIZATION=auth_header,
    )
    resp = server.create_token_response(request)
    assert resp.status_code == 400
    data = json.loads(resp.content)
    assert data["error"] == "invalid_request"
    assert "Missing" in data["error_description"]

    request = factory.post(
        "/oauth/token",
        data={"grant_type": "refresh_token", "refresh_token": "invalid"},
        HTTP_AUTHORIZATION=auth_header,
    )
    resp = server.create_token_response(request)
    assert resp.status_code == 400
    data = json.loads(resp.content)
    assert data["error"] == "invalid_grant"


def test_invalid_scope(factory, server, token):
    server.scopes_supported = ["profile"]
    request = factory.post(
        "/oauth/token",
        data={
            "grant_type": "refresh_token",
            "refresh_token": "r1",
            "scope": "invalid",
        },
        HTTP_AUTHORIZATION=create_basic_auth("client-id", "client-secret"),
    )
    resp = server.create_token_response(request)
    assert resp.status_code == 400
    data = json.loads(resp.content)
    assert data["error"] == "invalid_scope"


def test_authorize_tno_scope(factory, server, token):
    request = factory.post(
        "/oauth/token",
        data={
            "grant_type": "refresh_token",
            "refresh_token": "r1",
        },
        HTTP_AUTHORIZATION=create_basic_auth("client-id", "client-secret"),
    )
    resp = server.create_token_response(request)
    assert resp.status_code == 200
    data = json.loads(resp.content)
    assert "access_token" in data


def test_authorize_token_scope(factory, server, token):
    request = factory.post(
        "/oauth/token",
        data={
            "grant_type": "refresh_token",
            "refresh_token": "r1",
            "scope": "profile",
        },
        HTTP_AUTHORIZATION=create_basic_auth("client-id", "client-secret"),
    )
    resp = server.create_token_response(request)
    assert resp.status_code == 200
    data = json.loads(resp.content)
    assert "access_token" in data


def test_revoke_old_token(factory, server, token):
    request = factory.post(
        "/oauth/token",
        data={
            "grant_type": "refresh_token",
            "refresh_token": "r1",
            "scope": "profile",
        },
        HTTP_AUTHORIZATION=create_basic_auth("client-id", "client-secret"),
    )
    resp = server.create_token_response(request)
    assert resp.status_code == 200
    data = json.loads(resp.content)
    assert "access_token" in data

    resp = server.create_token_response(request)
    assert resp.status_code == 400



import pytest
from flask import json

from authlib.common.security import generate_token
from authlib.common.urls import url_decode
from authlib.common.urls import urlparse
from authlib.oauth2.rfc6749 import grants
from authlib.oauth2.rfc7636 import CodeChallenge as _CodeChallenge
from authlib.oauth2.rfc7636 import create_s256_code_challenge

from .models import CodeGrantMixin
from .models import save_authorization_code
from .oauth2_server import create_basic_header

authorize_url = "/oauth/authorize?response_type=code&client_id=client-id"


class AuthorizationCodeGrant(CodeGrantMixin, grants.AuthorizationCodeGrant):
    TOKEN_ENDPOINT_AUTH_METHODS = ["client_secret_basic", "client_secret_post", "none"]

    def save_authorization_code(self, code, request):
        return save_authorization_code(code, request)


class CodeChallenge(_CodeChallenge):
    SUPPORTED_CODE_CHALLENGE_METHOD = ["plain", "S256", "S128"]


@pytest.fixture(autouse=True)
def server(server):
    server.register_grant(AuthorizationCodeGrant, [CodeChallenge(required=True)])
    return server


@pytest.fixture(autouse=True)
def client(client, db):
    client.set_client_metadata(
        {
            "redirect_uris": ["https://client.test"],
            "scope": "profile address",
            "token_endpoint_auth_method": "none",
            "response_types": ["code"],
            "grant_types": ["authorization_code"],
        }
    )
    db.session.add(client)
    db.session.commit()
    return client


def test_missing_code_challenge(test_client):
    rv = test_client.get(authorize_url + "&code_challenge_method=plain")
    assert "Missing" in rv.location


def test_has_code_challenge(test_client):
    rv = test_client.get(
        authorize_url + "&code_challenge=Zhs2POMonIVVHZteWfoU7cSXQSm0YjghikFGJSDI2_s"
    )
    assert rv.data == b"ok"


def test_invalid_code_challenge(test_client):
    rv = test_client.get(
        authorize_url + "&code_challenge=abc&code_challenge_method=plain"
    )
    assert "Invalid" in rv.location


def test_invalid_code_challenge_method(test_client):
    suffix = "&code_challenge=Zhs2POMonIVVHZteWfoU7cSXQSm0YjghikFGJSDI2_s&code_challenge_method=invalid"
    rv = test_client.get(authorize_url + suffix)
    assert "Unsupported" in rv.location


def test_supported_code_challenge_method(test_client):
    suffix = "&code_challenge=Zhs2POMonIVVHZteWfoU7cSXQSm0YjghikFGJSDI2_s&code_challenge_method=plain"
    rv = test_client.get(authorize_url + suffix)
    assert rv.data == b"ok"


def test_trusted_client_without_code_challenge(test_client, db, client):
    client.client_secret = "client-secret"
    client.set_client_metadata(
        {
            "redirect_uris": ["https://client.test"],
            "scope": "profile address",
            "token_endpoint_auth_method": "client_secret_basic",
            "response_types": ["code"],
            "grant_types": ["authorization_code"],
        }
    )
    db.session.add(client)
    db.session.commit()

    rv = test_client.get(authorize_url)
    assert rv.data == b"ok"

    rv = test_client.post(authorize_url, data={"user_id": "1"})
    assert "code=" in rv.location

    params = dict(url_decode(urlparse.urlparse(rv.location).query))

    code = params["code"]
    headers = create_basic_header("client-id", "client-secret")
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "authorization_code",
            "code": code,
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert "access_token" in resp


def test_missing_code_verifier(test_client):
    url = authorize_url + "&code_challenge=Zhs2POMonIVVHZteWfoU7cSXQSm0YjghikFGJSDI2_s"
    rv = test_client.post(url, data={"user_id": "1"})
    assert "code=" in rv.location

    params = dict(url_decode(urlparse.urlparse(rv.location).query))
    code = params["code"]
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "authorization_code",
            "code": code,
            "client_id": "client-id",
        },
    )
    resp = json.loads(rv.data)
    assert "Missing" in resp["error_description"]


def test_trusted_client_missing_code_verifier(test_client, db, client):
    client.client_secret = "client-secret"
    client.set_client_metadata(
        {
            "redirect_uris": ["https://client.test"],
            "scope": "profile address",
            "token_endpoint_auth_method": "client_secret_basic",
            "response_types": ["code"],
            "grant_types": ["authorization_code"],
        }
    )
    db.session.add(client)
    db.session.commit()

    url = authorize_url + "&code_challenge=Zhs2POMonIVVHZteWfoU7cSXQSm0YjghikFGJSDI2_s"
    rv = test_client.post(url, data={"user_id": "1"})
    assert "code=" in rv.location

    params = dict(url_decode(urlparse.urlparse(rv.location).query))
    code = params["code"]
    headers = create_basic_header("client-id", "client-secret")
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "authorization_code",
            "code": code,
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert "Missing" in resp["error_description"]


def test_plain_code_challenge_invalid(test_client):
    url = authorize_url + "&code_challenge=Zhs2POMonIVVHZteWfoU7cSXQSm0YjghikFGJSDI2_s"
    rv = test_client.post(url, data={"user_id": "1"})
    assert "code=" in rv.location

    params = dict(url_decode(urlparse.urlparse(rv.location).query))
    code = params["code"]
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "authorization_code",
            "code": code,
            "code_verifier": "bar",
            "client_id": "client-id",
        },
    )
    resp = json.loads(rv.data)
    assert "Invalid" in resp["error_description"]


def test_plain_code_challenge_failed(test_client):
    url = authorize_url + "&code_challenge=Zhs2POMonIVVHZteWfoU7cSXQSm0YjghikFGJSDI2_s"
    rv = test_client.post(url, data={"user_id": "1"})
    assert "code=" in rv.location

    params = dict(url_decode(urlparse.urlparse(rv.location).query))
    code = params["code"]
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "authorization_code",
            "code": code,
            "code_verifier": generate_token(48),
            "client_id": "client-id",
        },
    )
    resp = json.loads(rv.data)
    assert "failed" in resp["error_description"]


def test_plain_code_challenge_success(test_client):
    code_verifier = generate_token(48)
    url = authorize_url + "&code_challenge=" + code_verifier
    rv = test_client.post(url, data={"user_id": "1"})
    assert "code=" in rv.location

    params = dict(url_decode(urlparse.urlparse(rv.location).query))
    code = params["code"]
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "authorization_code",
            "code": code,
            "code_verifier": code_verifier,
            "client_id": "client-id",
        },
    )
    resp = json.loads(rv.data)
    assert "access_token" in resp


def test_s256_code_challenge_success(test_client):
    code_verifier = generate_token(48)
    code_challenge = create_s256_code_challenge(code_verifier)
    url = authorize_url + "&code_challenge=" + code_challenge
    url += "&code_challenge_method=S256"

    rv = test_client.post(url, data={"user_id": "1"})
    assert "code=" in rv.location

    params = dict(url_decode(urlparse.urlparse(rv.location).query))
    code = params["code"]
    rv = test_client.post(
        "/oauth/token",
        data={
            "grant_type": "authorization_code",
            "code": code,
            "code_verifier": code_verifier,
            "client_id": "client-id",
        },
    )
    resp = json.loads(rv.data)
    assert "access_token" in resp


def test_not_implemented_code_challenge_method(test_client):
    url = authorize_url + "&code_challenge=Zhs2POMonIVVHZteWfoU7cSXQSm0YjghikFGJSDI2_s"
    url += "&code_challenge_method=S128"

    rv = test_client.post(url, data={"user_id": "1"})
    assert "code=" in rv.location

    params = dict(url_decode(urlparse.urlparse(rv.location).query))
    code = params["code"]
    with pytest.raises(RuntimeError):
        test_client.post(
            "/oauth/token",
            data={
                "grant_type": "authorization_code",
                "code": code,
                "code_verifier": generate_token(48),
                "client_id": "client-id",
            },
        )



import pytest

from authlib.oidc.discovery import OpenIDProviderMetadata
from authlib.oidc.discovery import get_well_known_url

WELL_KNOWN_URL = "/.well-known/openid-configuration"


def test_well_known_no_suffix_issuer():
    assert get_well_known_url("https://provider.test") == WELL_KNOWN_URL
    assert get_well_known_url("https://provider.test/") == WELL_KNOWN_URL


def test_well_known_with_suffix_issuer():
    assert (
        get_well_known_url("https://provider.test/issuer1")
        == "/issuer1" + WELL_KNOWN_URL
    )
    assert (
        get_well_known_url("https://provider.test/a/b/c") == "/a/b/c" + WELL_KNOWN_URL
    )


def test_well_known_with_external():
    assert (
        get_well_known_url("https://provider.test", external=True)
        == "https://provider.test" + WELL_KNOWN_URL
    )


def test_validate_jwks_uri():
    # required
    metadata = OpenIDProviderMetadata()
    with pytest.raises(ValueError, match='"jwks_uri" is required'):
        metadata.validate_jwks_uri()

    metadata = OpenIDProviderMetadata({"jwks_uri": "http://provider.test/jwks.json"})
    with pytest.raises(ValueError, match="https"):
        metadata.validate_jwks_uri()

    metadata = OpenIDProviderMetadata({"jwks_uri": "https://provider.test/jwks.json"})
    metadata.validate_jwks_uri()


def test_validate_acr_values_supported():
    _call_validate_array("acr_values_supported", ["urn:mace:incommon:iap:silver"])


def test_validate_subject_types_supported():
    _call_validate_array(
        "subject_types_supported", ["pairwise", "public"], required=True
    )
    _call_contains_invalid_value("subject_types_supported", ["invalid"])


def test_validate_id_token_signing_alg_values_supported():
    _call_validate_array(
        "id_token_signing_alg_values_supported",
        ["RS256"],
        required=True,
    )
    metadata = OpenIDProviderMetadata(
        {"id_token_signing_alg_values_supported": ["none"]}
    )
    with pytest.raises(ValueError, match="RS256"):
        metadata.validate_id_token_signing_alg_values_supported()


def test_validate_id_token_encryption_alg_values_supported():
    _call_validate_array("id_token_encryption_alg_values_supported", ["A128KW"])


def test_validate_id_token_encryption_enc_values_supported():
    _call_validate_array("id_token_encryption_enc_values_supported", ["A128GCM"])


def test_validate_userinfo_signing_alg_values_supported():
    _call_validate_array("userinfo_signing_alg_values_supported", ["RS256"])


def test_validate_userinfo_encryption_alg_values_supported():
    _call_validate_array("userinfo_encryption_alg_values_supported", ["A128KW"])


def test_validate_userinfo_encryption_enc_values_supported():
    _call_validate_array("userinfo_encryption_enc_values_supported", ["A128GCM"])


def test_validate_request_object_signing_alg_values_supported():
    _call_validate_array(
        "request_object_signing_alg_values_supported", ["none", "RS256"]
    )


def test_validate_request_object_encryption_alg_values_supported():
    _call_validate_array("request_object_encryption_alg_values_supported", ["A128KW"])


def test_validate_request_object_encryption_enc_values_supported():
    _call_validate_array("request_object_encryption_enc_values_supported", ["A128GCM"])


def test_validate_display_values_supported():
    _call_validate_array("display_values_supported", ["page", "touch"])
    _call_contains_invalid_value("display_values_supported", ["invalid"])


def test_validate_claim_types_supported():
    _call_validate_array("claim_types_supported", ["normal"])
    _call_contains_invalid_value("claim_types_supported", ["invalid"])
    metadata = OpenIDProviderMetadata()
    assert metadata.claim_types_supported == ["normal"]


def test_validate_claims_supported():
    _call_validate_array("claims_supported", ["sub"])


def test_validate_claims_locales_supported():
    _call_validate_array("claims_locales_supported", ["en-US"])


def test_validate_claims_parameter_supported():
    _call_validate_boolean("claims_parameter_supported")


def test_validate_request_parameter_supported():
    _call_validate_boolean("request_parameter_supported")


def test_validate_request_uri_parameter_supported():
    _call_validate_boolean("request_uri_parameter_supported", True)


def test_validate_require_request_uri_registration():
    _call_validate_boolean("require_request_uri_registration")


def _call_validate_boolean(key, default_value=False):
    def _validate(metadata):
        getattr(metadata, "validate_" + key)()

    metadata = OpenIDProviderMetadata()
    _validate(metadata)
    assert getattr(metadata, key) == default_value

    metadata = OpenIDProviderMetadata({key: "str"})
    with pytest.raises(ValueError, match="MUST be boolean"):
        _validate(metadata)

    metadata = OpenIDProviderMetadata({key: True})
    _validate(metadata)


def _call_validate_array(key, valid_value, required=False):
    def _validate(metadata):
        getattr(metadata, "validate_" + key)()

    metadata = OpenIDProviderMetadata()
    if required:
        with pytest.raises(ValueError, match=f'"{key}" is required'):
            _validate(metadata)

    else:
        _validate(metadata)

    # not array
    metadata = OpenIDProviderMetadata({key: "foo"})
    with pytest.raises(ValueError, match="JSON array"):
        _validate(metadata)

    # valid
    metadata = OpenIDProviderMetadata({key: valid_value})
    _validate(metadata)


def _call_contains_invalid_value(key, invalid_value):
    metadata = OpenIDProviderMetadata({key: invalid_value})
    with pytest.raises(ValueError, match=f'"{key}" contains invalid values'):
        getattr(metadata, "validate_" + key)()



import pytest
from flask import json

from authlib.integrations.sqla_oauth2 import create_revocation_endpoint

from .models import Client
from .models import Token
from .models import db
from .oauth2_server import create_basic_header


@pytest.fixture(autouse=True)
def server(server, app):
    RevocationEndpoint = create_revocation_endpoint(db.session, Token)
    server.register_endpoint(RevocationEndpoint)

    @app.route("/oauth/revoke", methods=["POST"])
    def revoke_token():
        return server.create_endpoint_response("revocation")

    return server


@pytest.fixture(autouse=True)
def client(client, db):
    client.set_client_metadata(
        {
            "scope": "profile",
            "redirect_uris": ["https://client.test/authorized"],
        }
    )
    db.session.add(client)
    db.session.commit()
    return client


def test_invalid_client(test_client):
    rv = test_client.post("/oauth/revoke")
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_client"

    headers = {"Authorization": "invalid token_string"}
    rv = test_client.post("/oauth/revoke", headers=headers)
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_client"

    headers = create_basic_header("invalid-client", "client-secret")
    rv = test_client.post("/oauth/revoke", headers=headers)
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_client"

    headers = create_basic_header("client-id", "invalid-secret")
    rv = test_client.post("/oauth/revoke", headers=headers)
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_client"


def test_invalid_token(test_client):
    headers = create_basic_header("client-id", "client-secret")
    rv = test_client.post("/oauth/revoke", headers=headers)
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_request"

    rv = test_client.post(
        "/oauth/revoke",
        data={
            "token": "invalid-token",
        },
        headers=headers,
    )
    assert rv.status_code == 200

    rv = test_client.post(
        "/oauth/revoke",
        data={
            "token": "a1",
            "token_type_hint": "unsupported_token_type",
        },
        headers=headers,
    )
    resp = json.loads(rv.data)
    assert resp["error"] == "unsupported_token_type"

    rv = test_client.post(
        "/oauth/revoke",
        data={
            "token": "a1",
            "token_type_hint": "refresh_token",
        },
        headers=headers,
    )
    assert rv.status_code == 200


def test_revoke_token_with_hint(test_client, token):
    headers = create_basic_header("client-id", "client-secret")
    rv = test_client.post(
        "/oauth/revoke",
        data={
            "token": "a1",
            "token_type_hint": "access_token",
        },
        headers=headers,
    )
    assert rv.status_code == 200


def test_revoke_token_without_hint(test_client, token):
    headers = create_basic_header("client-id", "client-secret")
    rv = test_client.post(
        "/oauth/revoke",
        data={
            "token": "a1",
        },
        headers=headers,
    )
    assert rv.status_code == 200


def test_revoke_token_bound_to_client(test_client, token):
    client2 = Client(
        user_id=1,
        client_id="client-id-2",
        client_secret="client-secret-2",
    )
    client2.set_client_metadata(
        {
            "scope": "profile",
            "redirect_uris": ["https://client.test/authorized"],
        }
    )
    db.session.add(client2)
    db.session.commit()

    headers = create_basic_header("client-id-2", "client-secret-2")
    rv = test_client.post(
        "/oauth/revoke",
        data={
            "token": "a1",
        },
        headers=headers,
    )
    assert rv.status_code == 400
    resp = json.loads(rv.data)
    assert resp["error"] == "invalid_grant"



import pytest

from tests.util import decode_response

from .oauth1_server import Client
from .oauth1_server import User
from .oauth1_server import create_authorization_server


@pytest.fixture(autouse=True)
def user(db):
    user = User(username="foo")
    db.session.add(user)
    db.session.commit()
    yield user
    db.session.delete(user)


@pytest.fixture(autouse=True)
def client(db, user):
    client = Client(
        user_id=user.id,
        client_id="client",
        client_secret="secret",
        default_redirect_uri="https://client.test",
    )
    db.session.add(client)
    db.session.commit()
    yield client
    db.session.delete(client)


@pytest.mark.parametrize("use_cache", [True, False])
def test_invalid_authorization(app, test_client, use_cache):
    create_authorization_server(app, use_cache, use_cache)
    url = "/oauth/authorize"

    # case 1
    rv = test_client.post(url, data={"user_id": "1"})
    data = decode_response(rv.data)
    assert data["error"] == "missing_required_parameter"
    assert "oauth_token" in data["error_description"]

    # case 2
    rv = test_client.post(url, data={"user_id": "1", "oauth_token": "a"})
    data = decode_response(rv.data)
    assert data["error"] == "invalid_token"


@pytest.mark.parametrize("use_cache", [True, False])
def test_authorize_denied(app, test_client, use_cache):
    create_authorization_server(app, use_cache, use_cache)
    initiate_url = "/oauth/initiate"
    authorize_url = "/oauth/authorize"

    rv = test_client.post(
        initiate_url,
        data={
            "oauth_consumer_key": "client",
            "oauth_callback": "oob",
            "oauth_signature_method": "PLAINTEXT",
            "oauth_signature": "secret&",
        },
    )
    data = decode_response(rv.data)
    assert "oauth_token" in data

    rv = test_client.post(authorize_url, data={"oauth_token": data["oauth_token"]})
    assert rv.status_code == 302
    assert "access_denied" in rv.headers["Location"]
    assert "https://client.test" in rv.headers["Location"]

    rv = test_client.post(
        initiate_url,
        data={
            "oauth_consumer_key": "client",
            "oauth_callback": "https://i.test",
            "oauth_signature_method": "PLAINTEXT",
            "oauth_signature": "secret&",
        },
    )
    data = decode_response(rv.data)
    assert "oauth_token" in data

    rv = test_client.post(authorize_url, data={"oauth_token": data["oauth_token"]})
    assert rv.status_code == 302
    assert "access_denied" in rv.headers["Location"]
    assert "https://i.test" in rv.headers["Location"]


@pytest.mark.parametrize("use_cache", [True, False])
def test_authorize_granted(app, test_client, use_cache):
    create_authorization_server(app, use_cache, use_cache)
    initiate_url = "/oauth/initiate"
    authorize_url = "/oauth/authorize"

    rv = test_client.post(
        initiate_url,
        data={
            "oauth_consumer_key": "client",
            "oauth_callback": "oob",
            "oauth_signature_method": "PLAINTEXT",
            "oauth_signature": "secret&",
        },
    )
    data = decode_response(rv.data)
    assert "oauth_token" in data

    rv = test_client.post(
        authorize_url, data={"user_id": "1", "oauth_token": data["oauth_token"]}
    )
    assert rv.status_code == 302
    assert "oauth_verifier" in rv.headers["Location"]
    assert "https://client.test" in rv.headers["Location"]

    rv = test_client.post(
        initiate_url,
        data={
            "oauth_consumer_key": "client",
            "oauth_callback": "https://i.test",
            "oauth_signature_method": "PLAINTEXT",
            "oauth_signature": "secret&",
        },
    )
    data = decode_response(rv.data)
    assert "oauth_token" in data

    rv = test_client.post(
        authorize_url, data={"user_id": "1", "oauth_token": data["oauth_token"]}
    )
    assert rv.status_code == 302
    assert "oauth_verifier" in rv.headers["Location"]
    assert "https://i.test" in rv.headers["Location"]



import hashlib

from authlib.common.encoding import to_bytes
from authlib.common.encoding import urlsafe_b64encode


def create_half_hash(s, alg):
    if alg == "EdDSA":
        hash_alg = hashlib.sha512
    else:
        hash_type = f"sha{alg[2:]}"
        hash_alg = getattr(hashlib, hash_type, None)
        if not hash_alg:
            return None

    data_digest = hash_alg(to_bytes(s)).digest()
    slice_index = int(len(data_digest) / 2)
    return urlsafe_b64encode(data_digest[:slice_index])



from collections import defaultdict


class Hookable:
    _hooks = None

    def __init__(self):
        self._hooks = defaultdict(set)

    def register_hook(self, hook_type, hook):
        self._hooks[hook_type].add(hook)

    def execute_hook(self, hook_type, *args, **kwargs):
        for hook in self._hooks[hook_type]:
            hook(self, *args, **kwargs)


def hooked(func=None, before=None, after=None):
    """Execute hooks before and after the decorated method."""

    def decorator(func):
        before_name = before or f"before_{func.__name__}"
        after_name = after or f"after_{func.__name__}"

        def wrapper(self, *args, **kwargs):
            self.execute_hook(before_name, *args, **kwargs)
            result = func(self, *args, **kwargs)
            self.execute_hook(after_name, result)
            return result

        return wrapper

    # The decorator has been called without parenthesis
    if callable(func):
        return decorator(func)

    return decorator



import pytest

from authlib.oauth2.rfc7662 import IntrospectionToken


def test_client_id():
    token = IntrospectionToken()
    assert token.client_id is None
    assert token.get_client_id() is None

    token = IntrospectionToken({"client_id": "foo"})
    assert token.client_id == "foo"
    assert token.get_client_id() == "foo"


def test_scope():
    token = IntrospectionToken()
    assert token.scope is None
    assert token.get_scope() is None

    token = IntrospectionToken({"scope": "foo"})
    assert token.scope == "foo"
    assert token.get_scope() == "foo"


def test_expires_in():
    token = IntrospectionToken()
    assert token.get_expires_in() == 0


def test_expires_at():
    token = IntrospectionToken()
    assert token.exp is None
    assert token.get_expires_at() == 0

    token = IntrospectionToken({"exp": 3600})
    assert token.exp == 3600
    assert token.get_expires_at() == 3600


def test_all_attributes():
    # https://tools.ietf.org/html/rfc7662#section-2.2
    token = IntrospectionToken()
    assert token.active is None
    assert token.scope is None
    assert token.client_id is None
    assert token.username is None
    assert token.token_type is None
    assert token.exp is None
    assert token.iat is None
    assert token.nbf is None
    assert token.sub is None
    assert token.aud is None
    assert token.iss is None
    assert token.jti is None


def test_invalid_attr():
    token = IntrospectionToken()
    with pytest.raises(AttributeError):
        token.invalid  # noqa:B018



from livereload import Server
from livereload import shell

app = Server()
# app.watch("src", shell("make build-docs"), delay=2)
app.watch("docs", shell("make build-docs"), delay=2)
app.serve(root="build/_html")



import time

import pytest
from flask import json

from authlib.common.security import generate_token
from authlib.jose import jwt
from authlib.oauth2.rfc6749.grants import (
    AuthorizationCodeGrant as _AuthorizationCodeGrant,
)
from authlib.oauth2.rfc7009 import RevocationEndpoint
from authlib.oauth2.rfc9068 import JWTRevocationEndpoint
from tests.util import read_file_path

from ..models import CodeGrantMixin
from ..models import User
from ..models import save_authorization_code
from ..oauth2_server import create_basic_header

issuer = "https://provider.test/"
resource_server = "resource-server-id"


@pytest.fixture
def jwks():
    return read_file_path("jwks_private.json")


@pytest.fixture(autouse=True)
def server(server):
    class AuthorizationCodeGrant(CodeGrantMixin, _AuthorizationCodeGrant):
        TOKEN_ENDPOINT_AUTH_METHODS = [
            "client_secret_basic",
            "client_secret_post",
            "none",
        ]

        def save_authorization_code(self, code, request):
            return save_authorization_code(code, request)

    server.register_grant(AuthorizationCodeGrant)
    return server


@pytest.fixture(autouse=True)
def revocation_endpoint(app, server, jwks):
    class MyJWTRevocationEndpoint(JWTRevocationEndpoint):
        def get_jwks(self):
            return jwks

    endpoint = MyJWTRevocationEndpoint(issuer=issuer)
    server.register_endpoint(endpoint)

    @app.route("/oauth/revoke", methods=["POST"])
    def revoke_token():
        return server.create_endpoint_response(MyJWTRevocationEndpoint.ENDPOINT_NAME)

    return endpoint


@pytest.fixture(autouse=True)
def user(db):
    user = User(username="foo")
    db.session.add(user)
    db.session.commit()
    yield user
    db.session.delete(user)


@pytest.fixture(autouse=True)
def client(client, db):
    client.set_client_metadata(
        {
            "scope": "profile",
            "redirect_uris": ["https://client.test/authorized"],
            "response_types": ["code"],
            "token_endpoint_auth_method": "client_secret_post",
            "grant_types": ["authorization_code"],
        }
    )
    db.session.add(client)
    db.session.commit()
    return client


def create_access_token_claims(client, user):
    now = int(time.time())
    expires_in = now + 3600
    auth_time = now - 60

    return {
        "iss": issuer,
        "exp": expires_in,
        "aud": [resource_server],
        "sub": user.get_user_id(),
        "client_id": client.client_id,
        "iat": now,
        "jti": generate_token(16),
        "auth_time": auth_time,
        "scope": client.scope,
        "groups": ["admins"],
        "roles": ["student"],
        "entitlements": ["captain"],
    }


@pytest.fixture
def claims(client, user):
    return create_access_token_claims(client, user)


def create_access_token(claims, jwks, alg="RS256", typ="at+jwt"):
    header = {"alg": alg, "typ": typ}
    access_token = jwt.encode(
        header,
        claims,
        key=jwks,
        check=False,
    )
    return access_token.decode()


@pytest.fixture
def access_token(claims, jwks):
    return create_access_token(claims, jwks)


def test_revocation(test_client, client, access_token):
    headers = create_basic_header(client.client_id, client.client_secret)
    rv = test_client.post(
        "/oauth/revoke", data={"token": access_token}, headers=headers
    )
    assert rv.status_code == 401
    resp = json.loads(rv.data)
    assert resp["error"] == "unsupported_token_type"


def test_non_access_token_skipped(test_client, server, client):
    class MyRevocationEndpoint(RevocationEndpoint):
        def query_token(self, token, token_type_hint):
            return None

    server.register_endpoint(MyRevocationEndpoint)
    headers = create_basic_header(client.client_id, client.client_secret)
    rv = test_client.post(
        "/oauth/revoke",
        data={
            "token": "refresh-token",
            "token_type_hint": "refresh_token",
        },
        headers=headers,
    )
    assert rv.status_code == 200
    resp = json.loads(rv.data)
    assert resp == {}


def test_access_token_non_jwt_skipped(test_client, server, client):
    class MyRevocationEndpoint(RevocationEndpoint):
        def query_token(self, token, token_type_hint):
            return None

    server.register_endpoint(MyRevocationEndpoint)
    headers = create_basic_header(client.client_id, client.client_secret)
    rv = test_client.post(
        "/oauth/revoke",
        data={
            "token": "non-jwt-access-token",
        },
        headers=headers,
    )
    assert rv.status_code == 200
    resp = json.loads(rv.data)
    assert resp == {}


def test_revocation_different_issuer(test_client, claims, jwks, client):
    claims["iss"] = "different-issuer"
    access_token = create_access_token(claims, jwks)

    headers = create_basic_header(client.client_id, client.client_secret)
    rv = test_client.post(
        "/oauth/revoke", data={"token": access_token}, headers=headers
    )
    assert rv.status_code == 401
    resp = json.loads(rv.data)
    assert resp["error"] == "unsupported_token_type"



import os
import struct

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.ciphers import Cipher
from cryptography.hazmat.primitives.ciphers.algorithms import AES
from cryptography.hazmat.primitives.ciphers.modes import GCM
from cryptography.hazmat.primitives.kdf.concatkdf import ConcatKDFHash
from cryptography.hazmat.primitives.keywrap import aes_key_unwrap
from cryptography.hazmat.primitives.keywrap import aes_key_wrap

from authlib.common.encoding import to_bytes
from authlib.common.encoding import to_native
from authlib.common.encoding import urlsafe_b64decode
from authlib.common.encoding import urlsafe_b64encode
from authlib.jose.rfc7516 import JWEAlgorithm

from .ec_key import ECKey
from .oct_key import OctKey
from .rsa_key import RSAKey


class DirectAlgorithm(JWEAlgorithm):
    name = "dir"
    description = "Direct use of a shared symmetric key"

    def prepare_key(self, raw_data):
        return OctKey.import_key(raw_data)

    def generate_preset(self, enc_alg, key):
        return {}

    def wrap(self, enc_alg, headers, key, preset=None):
        cek = key.get_op_key("encrypt")
        if len(cek) * 8 != enc_alg.CEK_SIZE:
            raise ValueError('Invalid "cek" length')
        return {"ek": b"", "cek": cek}

    def unwrap(self, enc_alg, ek, headers, key):
        cek = key.get_op_key("decrypt")
        if len(cek) * 8 != enc_alg.CEK_SIZE:
            raise ValueError('Invalid "cek" length')
        return cek


class RSAAlgorithm(JWEAlgorithm):
    #: A key of size 2048 bits or larger MUST be used with these algorithms
    #: RSA1_5, RSA-OAEP, RSA-OAEP-256
    key_size = 2048

    def __init__(self, name, description, pad_fn):
        self.name = name
        self.description = description
        self.padding = pad_fn

    def prepare_key(self, raw_data):
        return RSAKey.import_key(raw_data)

    def generate_preset(self, enc_alg, key):
        cek = enc_alg.generate_cek()
        return {"cek": cek}

    def wrap(self, enc_alg, headers, key, preset=None):
        if preset and "cek" in preset:
            cek = preset["cek"]
        else:
            cek = enc_alg.generate_cek()

        op_key = key.get_op_key("wrapKey")
        if op_key.key_size < self.key_size:
            raise ValueError("A key of size 2048 bits or larger MUST be used")
        ek = op_key.encrypt(cek, self.padding)
        return {"ek": ek, "cek": cek}

    def unwrap(self, enc_alg, ek, headers, key):
        # it will raise ValueError if failed
        op_key = key.get_op_key("unwrapKey")
        cek = op_key.decrypt(ek, self.padding)
        if len(cek) * 8 != enc_alg.CEK_SIZE:
            raise ValueError('Invalid "cek" length')
        return cek


class AESAlgorithm(JWEAlgorithm):
    def __init__(self, key_size):
        self.name = f"A{key_size}KW"
        self.description = f"AES Key Wrap using {key_size}-bit key"
        self.key_size = key_size

    def prepare_key(self, raw_data):
        return OctKey.import_key(raw_data)

    def generate_preset(self, enc_alg, key):
        cek = enc_alg.generate_cek()
        return {"cek": cek}

    def _check_key(self, key):
        if len(key) * 8 != self.key_size:
            raise ValueError(f"A key of size {self.key_size} bits is required.")

    def wrap_cek(self, cek, key):
        op_key = key.get_op_key("wrapKey")
        self._check_key(op_key)
        ek = aes_key_wrap(op_key, cek, default_backend())
        return {"ek": ek, "cek": cek}

    def wrap(self, enc_alg, headers, key, preset=None):
        if preset and "cek" in preset:
            cek = preset["cek"]
        else:
            cek = enc_alg.generate_cek()
        return self.wrap_cek(cek, key)

    def unwrap(self, enc_alg, ek, headers, key):
        op_key = key.get_op_key("unwrapKey")
        self._check_key(op_key)
        cek = aes_key_unwrap(op_key, ek, default_backend())
        if len(cek) * 8 != enc_alg.CEK_SIZE:
            raise ValueError('Invalid "cek" length')
        return cek


class AESGCMAlgorithm(JWEAlgorithm):
    EXTRA_HEADERS = frozenset(["iv", "tag"])

    def __init__(self, key_size):
        self.name = f"A{key_size}GCMKW"
        self.description = f"Key wrapping with AES GCM using {key_size}-bit key"
        self.key_size = key_size

    def prepare_key(self, raw_data):
        return OctKey.import_key(raw_data)

    def generate_preset(self, enc_alg, key):
        cek = enc_alg.generate_cek()
        return {"cek": cek}

    def _check_key(self, key):
        if len(key) * 8 != self.key_size:
            raise ValueError(f"A key of size {self.key_size} bits is required.")

    def wrap(self, enc_alg, headers, key, preset=None):
        if preset and "cek" in preset:
            cek = preset["cek"]
        else:
            cek = enc_alg.generate_cek()

        op_key = key.get_op_key("wrapKey")
        self._check_key(op_key)

        #: https://tools.ietf.org/html/rfc7518#section-4.7.1.1
        #: The "iv" (initialization vector) Header Parameter value is the
        #: base64url-encoded representation of the 96-bit IV value
        iv_size = 96
        iv = os.urandom(iv_size // 8)

        cipher = Cipher(AES(op_key), GCM(iv), backend=default_backend())
        enc = cipher.encryptor()
        ek = enc.update(cek) + enc.finalize()

        h = {
            "iv": to_native(urlsafe_b64encode(iv)),
            "tag": to_native(urlsafe_b64encode(enc.tag)),
        }
        return {"ek": ek, "cek": cek, "header": h}

    def unwrap(self, enc_alg, ek, headers, key):
        op_key = key.get_op_key("unwrapKey")
        self._check_key(op_key)

        iv = headers.get("iv")
        if not iv:
            raise ValueError('Missing "iv" in headers')

        tag = headers.get("tag")
        if not tag:
            raise ValueError('Missing "tag" in headers')

        iv = urlsafe_b64decode(to_bytes(iv))
        tag = urlsafe_b64decode(to_bytes(tag))

        cipher = Cipher(AES(op_key), GCM(iv, tag), backend=default_backend())
        d = cipher.decryptor()
        cek = d.update(ek) + d.finalize()
        if len(cek) * 8 != enc_alg.CEK_SIZE:
            raise ValueError('Invalid "cek" length')
        return cek


class ECDHESAlgorithm(JWEAlgorithm):
    EXTRA_HEADERS = ["epk", "apu", "apv"]
    ALLOWED_KEY_CLS = ECKey

    # https://tools.ietf.org/html/rfc7518#section-4.6
    def __init__(self, key_size=None):
        if key_size is None:
            self.name = "ECDH-ES"
            self.description = "ECDH-ES in the Direct Key Agreement mode"
        else:
            self.name = f"ECDH-ES+A{key_size}KW"
            self.description = (
                f"ECDH-ES using Concat KDF and CEK wrapped with A{key_size}KW"
            )
        self.key_size = key_size
        self.aeskw = AESAlgorithm(key_size)

    def prepare_key(self, raw_data):
        if isinstance(raw_data, self.ALLOWED_KEY_CLS):
            return raw_data
        return ECKey.import_key(raw_data)

    def generate_preset(self, enc_alg, key):
        epk = self._generate_ephemeral_key(key)
        h = self._prepare_headers(epk)
        preset = {"epk": epk, "header": h}
        if self.key_size is not None:
            cek = enc_alg.generate_cek()
            preset["cek"] = cek
        return preset

    def compute_fixed_info(self, headers, bit_size):
        # AlgorithmID
        if self.key_size is None:
            alg_id = u32be_len_input(headers["enc"])
        else:
            alg_id = u32be_len_input(headers["alg"])

        # PartyUInfo
        apu_info = u32be_len_input(headers.get("apu"), True)

        # PartyVInfo
        apv_info = u32be_len_input(headers.get("apv"), True)

        # SuppPubInfo
        pub_info = struct.pack(">I", bit_size)

        return alg_id + apu_info + apv_info + pub_info

    def compute_derived_key(self, shared_key, fixed_info, bit_size):
        ckdf = ConcatKDFHash(
            algorithm=hashes.SHA256(),
            length=bit_size // 8,
            otherinfo=fixed_info,
            backend=default_backend(),
        )
        return ckdf.derive(shared_key)

    def deliver(self, key, pubkey, headers, bit_size):
        shared_key = key.exchange_shared_key(pubkey)
        fixed_info = self.compute_fixed_info(headers, bit_size)
        return self.compute_derived_key(shared_key, fixed_info, bit_size)

    def _generate_ephemeral_key(self, key):
        return key.generate_key(key["crv"], is_private=True)

    def _prepare_headers(self, epk):
        # REQUIRED_JSON_FIELDS contains only public fields
        pub_epk = {k: epk[k] for k in epk.REQUIRED_JSON_FIELDS}
        pub_epk["kty"] = epk.kty
        return {"epk": pub_epk}

    def wrap(self, enc_alg, headers, key, preset=None):
        if self.key_size is None:
            bit_size = enc_alg.CEK_SIZE
        else:
            bit_size = self.key_size

        if preset and "epk" in preset:
            epk = preset["epk"]
            h = {}
        else:
            epk = self._generate_ephemeral_key(key)
            h = self._prepare_headers(epk)

        public_key = key.get_op_key("wrapKey")
        dk = self.deliver(epk, public_key, headers, bit_size)

        if self.key_size is None:
            return {"ek": b"", "cek": dk, "header": h}

        if preset and "cek" in preset:
            preset_for_kw = {"cek": preset["cek"]}
        else:
            preset_for_kw = None

        kek = self.aeskw.prepare_key(dk)
        rv = self.aeskw.wrap(enc_alg, headers, kek, preset_for_kw)
        rv["header"] = h
        return rv

    def unwrap(self, enc_alg, ek, headers, key):
        if "epk" not in headers:
            raise ValueError('Missing "epk" in headers')

        if self.key_size is None:
            bit_size = enc_alg.CEK_SIZE
        else:
            bit_size = self.key_size

        epk = key.import_key(headers["epk"])
        public_key = epk.get_op_key("wrapKey")
        dk = self.deliver(key, public_key, headers, bit_size)

        if self.key_size is None:
            return dk

        kek = self.aeskw.prepare_key(dk)
        return self.aeskw.unwrap(enc_alg, ek, headers, kek)


def u32be_len_input(s, base64=False):
    if not s:
        return b"\x00\x00\x00\x00"
    if base64:
        s = urlsafe_b64decode(to_bytes(s))
    else:
        s = to_bytes(s)
    return struct.pack(">I", len(s)) + s


JWE_ALG_ALGORITHMS = [
    DirectAlgorithm(),  # dir
    RSAAlgorithm("RSA1_5", "RSAES-PKCS1-v1_5", padding.PKCS1v15()),
    RSAAlgorithm(
        "RSA-OAEP",
        "RSAES OAEP using default parameters",
        padding.OAEP(padding.MGF1(hashes.SHA1()), hashes.SHA1(), None),
    ),
    RSAAlgorithm(
        "RSA-OAEP-256",
        "RSAES OAEP using SHA-256 and MGF1 with SHA-256",
        padding.OAEP(padding.MGF1(hashes.SHA256()), hashes.SHA256(), None),
    ),
    AESAlgorithm(128),  # A128KW
    AESAlgorithm(192),  # A192KW
    AESAlgorithm(256),  # A256KW
    AESGCMAlgorithm(128),  # A128GCMKW
    AESGCMAlgorithm(192),  # A192GCMKW
    AESGCMAlgorithm(256),  # A256GCMKW
    ECDHESAlgorithm(None),  # ECDH-ES
    ECDHESAlgorithm(128),  # ECDH-ES+A128KW
    ECDHESAlgorithm(192),  # ECDH-ES+A192KW
    ECDHESAlgorithm(256),  # ECDH-ES+A256KW
]

# 'PBES2-HS256+A128KW': '',
# 'PBES2-HS384+A192KW': '',
# 'PBES2-HS512+A256KW': '',



from django.contrib.auth.models import User
from django.db.models import CASCADE
from django.db.models import CharField
from django.db.models import ForeignKey
from django.db.models import Model
from django.db.models import TextField

from tests.util import read_file_path


class Client(Model):
    user = ForeignKey(User, on_delete=CASCADE)
    client_id = CharField(max_length=48, unique=True, db_index=True)
    client_secret = CharField(max_length=48, blank=True)
    default_redirect_uri = TextField(blank=False, default="")

    def get_default_redirect_uri(self):
        return self.default_redirect_uri

    def get_client_secret(self):
        return self.client_secret

    def get_rsa_public_key(self):
        return read_file_path("rsa_public.pem")


class TokenCredential(Model):
    user = ForeignKey(User, on_delete=CASCADE)
    client_id = CharField(max_length=48, db_index=True)
    oauth_token = CharField(max_length=84, unique=True, db_index=True)
    oauth_token_secret = CharField(max_length=84)

    def get_oauth_token(self):
        return self.oauth_token

    def get_oauth_token_secret(self):
        return self.oauth_token_secret



from io import StringIO
from unittest import mock

import pytest
import requests

from authlib.common.encoding import to_unicode
from authlib.integrations.requests_client import OAuth1Session
from authlib.integrations.requests_client import OAuthError
from authlib.oauth1 import SIGNATURE_PLAINTEXT
from authlib.oauth1 import SIGNATURE_RSA_SHA1
from authlib.oauth1 import SIGNATURE_TYPE_BODY
from authlib.oauth1 import SIGNATURE_TYPE_QUERY
from authlib.oauth1.rfc5849.util import escape

from ..util import mock_text_response
from ..util import read_key_file

TEST_RSA_OAUTH_SIGNATURE = (
    "Pko%2BFb4T1XGDE5DlLjuEMthVXjczqGi8qyfQ%2FSE405bBLEywint1tYNGN1me8h"
    "JoXZMqyXy%2F%2FAzJ0ViRYRc7rDTaTYyjB%2Fct%2FFt8f4lb3e9LfGhgkwih%2FsH2w%3D%3D"
)


def test_no_client_id():
    with pytest.raises(ValueError):
        OAuth1Session(None)


def test_signature_types():
    def verify_signature(getter):
        def fake_send(r, **kwargs):
            signature = to_unicode(getter(r))
            assert "oauth_signature" in signature
            resp = mock.MagicMock(spec=requests.Response)
            resp.cookies = []
            return resp

        return fake_send

    header = OAuth1Session("foo")
    header.send = verify_signature(lambda r: r.headers["Authorization"])
    header.post("https://provider.test")

    query = OAuth1Session("foo", signature_type=SIGNATURE_TYPE_QUERY)
    query.send = verify_signature(lambda r: r.url)
    query.post("https://provider.test")

    body = OAuth1Session("foo", signature_type=SIGNATURE_TYPE_BODY)
    headers = {"Content-Type": "application/x-www-form-urlencoded"}
    body.send = verify_signature(lambda r: r.body)
    body.post("https://provider.test", headers=headers, data="")


@mock.patch("authlib.oauth1.rfc5849.client_auth.generate_timestamp")
@mock.patch("authlib.oauth1.rfc5849.client_auth.generate_nonce")
def test_signature_methods(generate_nonce, generate_timestamp):
    generate_nonce.return_value = "abc"
    generate_timestamp.return_value = "123"

    signature = ", ".join(
        [
            'OAuth oauth_nonce="abc"',
            'oauth_timestamp="123"',
            'oauth_version="1.0"',
            'oauth_signature_method="HMAC-SHA1"',
            'oauth_consumer_key="foo"',
            'oauth_signature="GuqiSr5%2FHajrrmc%2FFprUV4cCGbw%3D"',
        ]
    )
    auth = OAuth1Session("foo")
    auth.send = verify_signature(signature)
    auth.post("https://provider.test")

    signature = (
        "OAuth "
        'oauth_nonce="abc", oauth_timestamp="123", oauth_version="1.0", '
        'oauth_signature_method="PLAINTEXT", oauth_consumer_key="foo", '
        'oauth_signature="%26"'
    )
    auth = OAuth1Session("foo", signature_method=SIGNATURE_PLAINTEXT)
    auth.send = verify_signature(signature)
    auth.post("https://provider.test")

    signature = (
        "OAuth "
        'oauth_nonce="abc", oauth_timestamp="123", oauth_version="1.0", '
        'oauth_signature_method="RSA-SHA1", oauth_consumer_key="foo", '
        f'oauth_signature="{TEST_RSA_OAUTH_SIGNATURE}"'
    )

    rsa_key = read_key_file("rsa_private.pem")
    auth = OAuth1Session("foo", signature_method=SIGNATURE_RSA_SHA1, rsa_key=rsa_key)
    auth.send = verify_signature(signature)
    auth.post("https://provider.test")


@mock.patch("authlib.oauth1.rfc5849.client_auth.generate_timestamp")
@mock.patch("authlib.oauth1.rfc5849.client_auth.generate_nonce")
def test_binary_upload(generate_nonce, generate_timestamp):
    generate_nonce.return_value = "abc"
    generate_timestamp.return_value = "123"
    fake_xml = StringIO("hello world")
    headers = {"Content-Type": "application/xml"}

    def fake_send(r, **kwargs):
        auth_header = r.headers["Authorization"]
        assert "oauth_body_hash" in auth_header

    auth = OAuth1Session("foo", force_include_body=True)
    auth.send = fake_send
    auth.post("https://provider.test", headers=headers, files=[("fake", fake_xml)])


@mock.patch("authlib.oauth1.rfc5849.client_auth.generate_timestamp")
@mock.patch("authlib.oauth1.rfc5849.client_auth.generate_nonce")
def test_nonascii(generate_nonce, generate_timestamp):
    generate_nonce.return_value = "abc"
    generate_timestamp.return_value = "123"
    signature = (
        'OAuth oauth_nonce="abc", oauth_timestamp="123", oauth_version="1.0", '
        'oauth_signature_method="HMAC-SHA1", oauth_consumer_key="foo", '
        'oauth_signature="USkqQvV76SCKBewYI9cut6FfYcI%3D"'
    )
    auth = OAuth1Session("foo")
    auth.send = verify_signature(signature)
    auth.post("https://provider.test?cjk=%E5%95%A6%E5%95%A6")


def test_redirect_uri():
    sess = OAuth1Session("foo")
    assert sess.redirect_uri is None
    url = "https://provider.test"
    sess.redirect_uri = url
    assert sess.redirect_uri == url


def test_set_token():
    sess = OAuth1Session("foo")
    try:
        sess.token = {}
    except OAuthError as exc:
        assert exc.error == "missing_token"

    sess.token = {"oauth_token": "a", "oauth_token_secret": "b"}
    assert sess.token["oauth_verifier"] is None
    sess.token = {"oauth_token": "a", "oauth_verifier": "c"}
    assert sess.token["oauth_token_secret"] == "b"
    assert sess.token["oauth_verifier"] == "c"

    sess.token = None
    assert sess.token["oauth_token"] is None
    assert sess.token["oauth_token_secret"] is None
    assert sess.token["oauth_verifier"] is None


def test_create_authorization_url():
    auth = OAuth1Session("foo")
    url = "https://provider.test/authorize"
    token = "asluif023sf"
    auth_url = auth.create_authorization_url(url, request_token=token)
    assert auth_url == url + "?oauth_token=" + token
    redirect_uri = "https://client.test/callback"
    auth = OAuth1Session("foo", redirect_uri=redirect_uri)
    auth_url = auth.create_authorization_url(url, request_token=token)
    assert escape(redirect_uri) in auth_url


def test_parse_response_url():
    url = "https://provider.test/callback?oauth_token=foo&oauth_verifier=bar"
    auth = OAuth1Session("foo")
    resp = auth.parse_authorization_response(url)
    assert resp["oauth_token"] == "foo"
    assert resp["oauth_verifier"] == "bar"
    for k, v in resp.items():
        assert isinstance(k, str)
        assert isinstance(v, str)


def test_fetch_request_token():
    auth = OAuth1Session("foo", realm="A")
    auth.send = mock_text_response("oauth_token=foo")
    resp = auth.fetch_request_token("https://provider.test/token")
    assert resp["oauth_token"] == "foo"
    for k, v in resp.items():
        assert isinstance(k, str)
        assert isinstance(v, str)

    resp = auth.fetch_request_token("https://provider.test/token")
    assert resp["oauth_token"] == "foo"


def test_fetch_request_token_with_optional_arguments():
    auth = OAuth1Session("foo")
    auth.send = mock_text_response("oauth_token=foo")
    resp = auth.fetch_request_token(
        "https://provider.test/token", verify=False, stream=True
    )
    assert resp["oauth_token"] == "foo"
    for k, v in resp.items():
        assert isinstance(k, str)
        assert isinstance(v, str)


def test_fetch_access_token():
    auth = OAuth1Session("foo", verifier="bar")
    auth.send = mock_text_response("oauth_token=foo")
    resp = auth.fetch_access_token("https://provider.test/token")
    assert resp["oauth_token"] == "foo"
    for k, v in resp.items():
        assert isinstance(k, str)
        assert isinstance(v, str)

    auth = OAuth1Session("foo", verifier="bar")
    auth.send = mock_text_response('{"oauth_token":"foo"}')
    resp = auth.fetch_access_token("https://provider.test/token")
    assert resp["oauth_token"] == "foo"

    auth = OAuth1Session("foo")
    auth.send = mock_text_response("oauth_token=foo")
    resp = auth.fetch_access_token("https://provider.test/token", verifier="bar")
    assert resp["oauth_token"] == "foo"


def test_fetch_access_token_with_optional_arguments():
    auth = OAuth1Session("foo", verifier="bar")
    auth.send = mock_text_response("oauth_token=foo")
    resp = auth.fetch_access_token(
        "https://provider.test/token", verify=False, stream=True
    )
    assert resp["oauth_token"] == "foo"
    for k, v in resp.items():
        assert isinstance(k, str)
        assert isinstance(v, str)


def _test_fetch_access_token_raises_error(session):
    """Assert that an error is being raised whenever there's no verifier
    passed in to the client.
    """
    session.send = mock_text_response("oauth_token=foo")
    with pytest.raises(OAuthError, match="missing_verifier"):
        session.fetch_access_token("https://provider.test/token")


def test_fetch_token_invalid_response():
    auth = OAuth1Session("foo")
    auth.send = mock_text_response("not valid urlencoded response!")
    with pytest.raises(ValueError):
        auth.fetch_request_token("https://provider.test/token")

    for code in (400, 401, 403):
        auth.send = mock_text_response("valid=response", code)
        with pytest.raises(OAuthError, match="fetch_token_denied"):
            auth.fetch_request_token("https://provider.test/token")


def test_fetch_access_token_missing_verifier():
    _test_fetch_access_token_raises_error(OAuth1Session("foo"))


def test_fetch_access_token_has_verifier_is_none():
    session = OAuth1Session("foo")
    session.auth.verifier = None
    _test_fetch_access_token_raises_error(session)


def verify_signature(signature):
    def fake_send(r, **kwargs):
        auth_header = to_unicode(r.headers["Authorization"])
        # RSA signatures are non-deterministic, so we only check the prefix for RSA-SHA1
        if 'oauth_signature_method="RSA-SHA1"' in signature:
            signature_prefix = (
                signature.split('oauth_signature="')[0] + 'oauth_signature="'
            )
            auth_prefix = (
                auth_header.split('oauth_signature="')[0] + 'oauth_signature="'
            )
            assert auth_prefix == signature_prefix
        else:
            assert auth_header == signature
        resp = mock.MagicMock(spec=requests.Response)
        resp.cookies = []
        return resp

    return fake_send



"""authlib.oauth2.rfc8414.
~~~~~~~~~~~~~~~~~~~~~~

This module represents a direct implementation of
OAuth 2.0 Authorization Server Metadata.

https://tools.ietf.org/html/rfc8414
"""

from .models import AuthorizationServerMetadata
from .well_known import get_well_known_url

__all__ = ["AuthorizationServerMetadata", "get_well_known_url"]



import base64

import pytest

from authlib.oauth2.rfc6749 import errors
from authlib.oauth2.rfc6749 import parameters
from authlib.oauth2.rfc6749 import util


def test_parse_authorization_code_response():
    with pytest.raises(errors.MissingCodeException):
        parameters.parse_authorization_code_response(
            "https://provider.test/?state=c",
        )

    with pytest.raises(errors.MismatchingStateException):
        parameters.parse_authorization_code_response(
            "https://provider.test/?code=a&state=c",
            "b",
        )

    url = "https://provider.test/?code=a&state=c"
    rv = parameters.parse_authorization_code_response(url, "c")
    assert rv == {"code": "a", "state": "c"}


def test_parse_implicit_response():
    with pytest.raises(errors.MissingTokenException):
        parameters.parse_implicit_response(
            "https://provider.test/#a=b",
        )

    with pytest.raises(errors.MissingTokenTypeException):
        parameters.parse_implicit_response(
            "https://provider.test/#access_token=a",
        )

    with pytest.raises(errors.MismatchingStateException):
        parameters.parse_implicit_response(
            "https://provider.test/#access_token=a&token_type=bearer&state=c",
            "abc",
        )

    url = "https://provider.test/#access_token=a&token_type=bearer&state=c"
    rv = parameters.parse_implicit_response(url, "c")
    assert rv == {"access_token": "a", "token_type": "bearer", "state": "c"}


def test_prepare_grant_uri():
    grant_uri = parameters.prepare_grant_uri(
        "https://provider.test/authorize", "dev", "code", max_age=0, resource=["a", "b"]
    )
    assert (
        grant_uri
        == "https://provider.test/authorize?response_type=code&client_id=dev&max_age=0&resource=a&resource=b"
    )


def test_list_to_scope():
    assert util.list_to_scope(["a", "b"]) == "a b"
    assert util.list_to_scope("a b") == "a b"
    assert util.list_to_scope(None) is None


def test_scope_to_list():
    assert util.scope_to_list("a b") == ["a", "b"]
    assert util.scope_to_list(["a", "b"]) == ["a", "b"]
    assert util.scope_to_list(None) is None


def test_extract_basic_authorization():
    assert util.extract_basic_authorization({}) == (None, None)
    assert util.extract_basic_authorization({"Authorization": "invalid"}) == (
        None,
        None,
    )

    text = "Basic invalid-base64"
    assert util.extract_basic_authorization({"Authorization": text}) == (None, None)

    text = "Basic {}".format(base64.b64encode(b"a").decode())
    assert util.extract_basic_authorization({"Authorization": text}) == ("a", None)

    text = "Basic {}".format(base64.b64encode(b"a:b").decode())
    assert util.extract_basic_authorization({"Authorization": text}) == ("a", "b")



import json
import os
import time
from unittest import mock

import requests

ROOT = os.path.abspath(os.path.dirname(__file__))


def read_key_file(name):
    file_path = os.path.join(ROOT, "keys", name)
    with open(file_path) as f:
        if name.endswith(".json"):
            return json.load(f)
        return f.read()


def mock_text_response(body, status_code=200):
    def fake_send(r, **kwargs):
        resp = mock.MagicMock(spec=requests.Response)
        resp.cookies = []
        resp.text = body
        resp.status_code = status_code
        return resp

    return fake_send


def mock_send_value(body, status_code=200):
    resp = mock.MagicMock(spec=requests.Response)
    resp.cookies = []
    if isinstance(body, dict):
        resp.json = lambda: body
    else:
        resp.text = body
    resp.status_code = status_code
    return resp


def get_bearer_token():
    return {
        "token_type": "Bearer",
        "access_token": "a",
        "refresh_token": "b",
        "expires_in": "3600",
        "expires_at": int(time.time()) + 3600,
    }



import time

import pytest
from httpx import ASGITransport

from authlib.integrations.httpx_client import AsyncAssertionClient

from ..asgi_helper import AsyncMockDispatch

default_token = {
    "token_type": "Bearer",
    "access_token": "a",
    "refresh_token": "b",
    "expires_in": "3600",
    "expires_at": int(time.time()) + 3600,
}


@pytest.mark.asyncio
async def test_refresh_token():
    async def verifier(request):
        content = await request.body()
        if str(request.url) == "https://provider.test/token":
            assert b"assertion=" in content

    async with AsyncAssertionClient(
        "https://provider.test/token",
        grant_type=AsyncAssertionClient.JWT_BEARER_GRANT_TYPE,
        issuer="foo",
        subject="foo",
        audience="foo",
        alg="HS256",
        key="secret",
        transport=ASGITransport(AsyncMockDispatch(default_token, assert_func=verifier)),
    ) as client:
        await client.get("https://provider.test")

    # trigger more case
    now = int(time.time())
    async with AsyncAssertionClient(
        "https://provider.test/token",
        issuer="foo",
        subject=None,
        audience="foo",
        issued_at=now,
        expires_at=now + 3600,
        header={"alg": "HS256"},
        key="secret",
        scope="email",
        claims={"test_mode": "true"},
        transport=ASGITransport(AsyncMockDispatch(default_token, assert_func=verifier)),
    ) as client:
        await client.get("https://provider.test")
        await client.get("https://provider.test")


@pytest.mark.asyncio
async def test_without_alg():
    async with AsyncAssertionClient(
        "https://provider.test/token",
        issuer="foo",
        subject="foo",
        audience="foo",
        key="secret",
        transport=ASGITransport(AsyncMockDispatch()),
    ) as client:
        with pytest.raises(ValueError):
            await client.get("https://provider.test")



import json

import pytest
from django.http import JsonResponse

from authlib.integrations.django_oauth2 import BearerTokenValidator
from authlib.integrations.django_oauth2 import ResourceProtector

from .models import Client
from .models import OAuth2Token

require_oauth = ResourceProtector()
require_oauth.register_token_validator(BearerTokenValidator(OAuth2Token))


@pytest.fixture(autouse=True)
def client(user):
    client = Client(
        user_id=user.pk,
        client_id="client-id",
        client_secret="client-secret",
        scope="profile",
    )
    client.save()
    yield client
    client.delete()


def test_invalid_token(factory):
    @require_oauth("profile")
    def get_user_profile(request):
        user = request.oauth_token.user
        return JsonResponse(dict(sub=user.pk, username=user.username))

    request = factory.get("/user")
    resp = get_user_profile(request)
    assert resp.status_code == 401
    data = json.loads(resp.content)
    assert data["error"] == "missing_authorization"

    request = factory.get("/user", HTTP_AUTHORIZATION="invalid token")
    resp = get_user_profile(request)
    assert resp.status_code == 401
    data = json.loads(resp.content)
    assert data["error"] == "unsupported_token_type"

    request = factory.get("/user", HTTP_AUTHORIZATION="bearer token")
    resp = get_user_profile(request)
    assert resp.status_code == 401
    data = json.loads(resp.content)
    assert data["error"] == "invalid_token"


def test_expired_token(factory, token):
    token.expires_in = -10
    token.save()

    @require_oauth("profile")
    def get_user_profile(request):
        user = request.oauth_token.user
        return JsonResponse(dict(sub=user.pk, username=user.username))

    request = factory.get("/user", HTTP_AUTHORIZATION="bearer a1")
    resp = get_user_profile(request)
    assert resp.status_code == 401
    data = json.loads(resp.content)
    assert data["error"] == "invalid_token"


def test_insufficient_token(factory, token):
    @require_oauth("email")
    def get_user_email(request):
        user = request.oauth_token.user
        return JsonResponse(dict(email=user.email))

    request = factory.get("/user/email", HTTP_AUTHORIZATION="bearer a1")
    resp = get_user_email(request)
    assert resp.status_code == 403
    data = json.loads(resp.content)
    assert data["error"] == "insufficient_scope"


def test_access_resource(factory, token):
    @require_oauth("profile", optional=True)
    def get_user_profile(request):
        if request.oauth_token:
            user = request.oauth_token.user
            return JsonResponse(dict(sub=user.pk, username=user.username))
        return JsonResponse(dict(sub=0, username="anonymous"))

    request = factory.get("/user")
    resp = get_user_profile(request)
    assert resp.status_code == 200
    data = json.loads(resp.content)
    assert data["username"] == "anonymous"

    request = factory.get("/user", HTTP_AUTHORIZATION="bearer a1")
    resp = get_user_profile(request)
    assert resp.status_code == 200
    data = json.loads(resp.content)
    assert data["username"] == "foo"


def test_scope_operator(factory, token):
    @require_oauth(["profile email"])
    def operator_and(request):
        user = request.oauth_token.user
        return JsonResponse(dict(sub=user.pk, username=user.username))

    @require_oauth(["profile", "email"])
    def operator_or(request):
        user = request.oauth_token.user
        return JsonResponse(dict(sub=user.pk, username=user.username))

    request = factory.get("/user", HTTP_AUTHORIZATION="bearer a1")
    resp = operator_and(request)
    assert resp.status_code == 403
    data = json.loads(resp.content)
    assert data["error"] == "insufficient_scope"

    resp = operator_or(request)
    assert resp.status_code == 200
    data = json.loads(resp.content)
    assert data["username"] == "foo"



import pytest

from authlib.jose.errors import InvalidClaimError
from authlib.jose.errors import MissingClaimError
from authlib.oidc.core import CodeIDToken
from authlib.oidc.core import HybridIDToken
from authlib.oidc.core import ImplicitIDToken
from authlib.oidc.core import UserInfo
from authlib.oidc.core import get_claim_cls_by_response_type


def test_essential_claims():
    claims = CodeIDToken({}, {})
    with pytest.raises(MissingClaimError):
        claims.validate()
    claims = CodeIDToken(
        {"iss": "1", "sub": "1", "aud": "1", "exp": 10000, "iat": 100}, {}
    )
    claims.validate(1000)


def test_validate_auth_time():
    claims = CodeIDToken(
        {"iss": "1", "sub": "1", "aud": "1", "exp": 10000, "iat": 100}, {}
    )
    claims.params = {"max_age": 100}
    with pytest.raises(MissingClaimError):
        claims.validate(1000)

    claims["auth_time"] = "foo"
    with pytest.raises(InvalidClaimError):
        claims.validate(1000)


def test_validate_nonce():
    claims = CodeIDToken(
        {"iss": "1", "sub": "1", "aud": "1", "exp": 10000, "iat": 100}, {}
    )
    claims.params = {"nonce": "foo"}
    with pytest.raises(MissingClaimError):
        claims.validate(1000)
    claims["nonce"] = "bar"
    with pytest.raises(InvalidClaimError):
        claims.validate(1000)
    claims["nonce"] = "foo"
    claims.validate(1000)


def test_validate_amr():
    claims = CodeIDToken(
        {
            "iss": "1",
            "sub": "1",
            "aud": "1",
            "exp": 10000,
            "iat": 100,
            "amr": "invalid",
        },
        {},
    )
    with pytest.raises(InvalidClaimError):
        claims.validate(1000)


def test_validate_azp():
    claims = CodeIDToken(
        {
            "iss": "1",
            "sub": "1",
            "aud": "1",
            "exp": 10000,
            "iat": 100,
        },
        {},
    )
    claims.params = {"client_id": "2"}
    with pytest.raises(MissingClaimError):
        claims.validate(1000)

    claims["azp"] = "1"
    with pytest.raises(InvalidClaimError):
        claims.validate(1000)

    claims["azp"] = "2"
    claims.validate(1000)


def test_validate_at_hash():
    claims = CodeIDToken(
        {
            "iss": "1",
            "sub": "1",
            "aud": "1",
            "exp": 10000,
            "iat": 100,
            "at_hash": "a",
        },
        {},
    )
    claims.params = {"access_token": "a"}

    # invalid alg won't raise
    claims.header = {"alg": "HS222"}
    claims.validate(1000)

    claims.header = {"alg": "HS256"}
    with pytest.raises(InvalidClaimError):
        claims.validate(1000)


def test_implicit_id_token():
    claims = ImplicitIDToken(
        {
            "iss": "1",
            "sub": "1",
            "aud": "1",
            "exp": 10000,
            "iat": 100,
            "nonce": "a",
        },
        {},
    )
    claims.params = {"access_token": "a"}
    with pytest.raises(MissingClaimError):
        claims.validate(1000)


def test_hybrid_id_token():
    claims = HybridIDToken(
        {
            "iss": "1",
            "sub": "1",
            "aud": "1",
            "exp": 10000,
            "iat": 100,
            "nonce": "a",
        },
        {},
    )
    claims.validate(1000)

    claims.params = {"code": "a"}
    with pytest.raises(MissingClaimError):
        claims.validate(1000)

    # invalid alg won't raise
    claims.header = {"alg": "HS222"}
    claims["c_hash"] = "a"
    claims.validate(1000)

    claims.header = {"alg": "HS256"}
    with pytest.raises(InvalidClaimError):
        claims.validate(1000)


def test_get_claim_cls_by_response_type():
    cls = get_claim_cls_by_response_type("id_token")
    assert cls == ImplicitIDToken
    cls = get_claim_cls_by_response_type("code")
    assert cls == CodeIDToken
    cls = get_claim_cls_by_response_type("code id_token")
    assert cls == HybridIDToken
    cls = get_claim_cls_by_response_type("none")
    assert cls is None


def test_userinfo_getattribute():
    user = UserInfo({"sub": "1"})
    assert user.sub == "1"
    assert user.email is None
    with pytest.raises(AttributeError):
        user.invalid  # noqa: B018
